"""FastAPI server."""

import os

import fastapi
import uvicorn
from fastapi import FastAPI, Response
from starlette import status

from lemmatizer_be import BnkorpusLemmatizer

lm = BnkorpusLemmatizer()
app = FastAPI()


@app.get("/lemmas")
def get_lemmas(word: str) -> Response:
    """Get lemmas for a word.

    Parameters
    ----------
    word : str
        query

    Returns
    -------
    Response
        response

    """
    return {"result": lm.lemmas(word)}


@app.get("/lemma")
def get_lemma(word: str) -> Response:
    """Get one (the shortest) lemma for a word.

    Parameters
    ----------
    word : str
        query

    Returns
    -------
    Response
        response

    """
    return {"result": lm.lemmatize(word)}


@app.get("/")
def redirect_to_docs() -> Response:
    """Redirect to docs when accessing /.

    Returns
    -------
    Response
        _description_

    """
    return fastapi.responses.RedirectResponse("/docs", status_code=status.HTTP_302_FOUND)


def main():  # noqa: D103
    uvicorn.run(
        app,
        host="localhost",
        port=int(os.environ.get("PORT", "8093")),
        log_level="info",
    )


if __name__ == "__main__":
    main()



"""Build lemma dictionary from bnkorpus."""

# ruff: noqa: T201

import os
import sqlite3
import sys
import zipfile
from pathlib import Path

from lxml import etree
from tqdm import tqdm

from lemmatizer_be._utils import _fetch_unzip, dir_empty

DATA_DIR = Path(
    os.environ.get(
        "LEMMATIZER_BE_DATA_DIR",
        Path(
            "~",
            ".alerus",
            "shared",
        ),
    )
).expanduser()
DATA_DIR.mkdir(parents=True, exist_ok=True)

BNKORPUS_DIR = Path("~", ".alerus", "shared", "bnkorpus").expanduser()
BNKORPUS_DIR.mkdir(parents=True, exist_ok=True)

BNKORPUS_URL = "https://github.com/Belarus/GrammarDB/releases/download/RELEASE-202309/RELEASE-20230920.zip"


def strip_plus(word):  # noqa: D103
    return word.replace("+", "")


def main():  # noqa: D103
    print("bnkorpus status:", end=" ")

    if dir_empty(BNKORPUS_DIR):
        print("missing. Downloading...")
        _fetch_unzip(BNKORPUS_URL, BNKORPUS_DIR)
    else:
        print("OK")

    # region Creating database
    db_path = str(DATA_DIR / "lemma_data.sqlite3")
    if Path(db_path).is_file():
        os.remove(db_path)

    connection = sqlite3.connect(db_path)
    cursor = connection.cursor()

    cursor.execute("""
        CREATE TABLE IF NOT EXISTS lemma_data (
            form TEXT NOT NULL PRIMARY KEY,
            lemmas TEXT NOT NULL
        );
    """)

    # cursor.execute("""
    #     CREATE INDEX IF NOT EXISTS index_forms ON lemma_data (form);
    #                """)

    print(f"Initialized empty db '{Path(db_path).resolve()}'")

    connection.commit()
    # endregion

    data_dict = {}

    for xml_path in BNKORPUS_DIR.glob("*.xml"):
        tree = etree.fromstring(xml_path.read_bytes())  # noqa: S320
        print(f"Loaded '{xml_path}'. Analyzing...", end=" ")
        sys.stdout.flush()

        for paradigm in tree.findall("Paradigm"):
            paradigm_lemma = strip_plus(paradigm.get("lemma"))

            for variant in paradigm.findall("Variant"):
                pos = paradigm.get("tag", variant.get("tag", "0"))[0]

                for form in variant.findall("Form"):
                    form_text = strip_plus(form.text)

                    if form_text not in data_dict:
                        data_dict[form_text] = set()

                    data_dict[form_text].add(paradigm_lemma + "|" + pos)

        print("OK")

    changeable = {}

    for k, v in data_dict.items():
        list_v = list(v)

        for i, val in enumerate(list_v):
            if val.split("|")[0] == k:  # If form == lemma then write |A, not word|A
                list_v[i] = "|" + list_v[0].split("|")[1]

        changeable[k] = sorted(list_v, key=len)

    print(f"Found {len(changeable):_} words")

    # Sort by forms ascending
    changeable = sorted(changeable.items(), key=lambda i: i[0])

    # region Writing data
    for word, lemmas in tqdm(changeable, desc="Writing the data into sqlite3 db..."):
        lemmas_val = ";".join(lemmas)

        sqlite_insert_with_param = """INSERT INTO lemma_data
                              (form, lemmas)
                              VALUES (?, ?);"""

        data_tuple = (word, lemmas_val)
        cursor.execute(sqlite_insert_with_param, data_tuple)

    connection.commit()

    print(f"The changeable db size is {(Path(db_path).stat().st_size / 1024 / 1024):.2f} MB")

    cursor.execute("""VACUUM;""")

    connection.close()
    # endregion

    # region Compressing
    arc_path = DATA_DIR / "lemma_data.zip"

    with zipfile.ZipFile(
        str(arc_path.resolve()),
        mode="w",
        compression=zipfile.ZIP_DEFLATED,
        compresslevel=6,
    ) as zip_file:
        zip_file.write(DATA_DIR / "lemma_data.sqlite3", "lemma_data.sqlite3")

    print(f"The arc file size is {(arc_path.stat().st_size / 1024 / 1024):.2f} MB")
    # endregion


if __name__ == "__main__":
    main()



"""The lemmatizer main file."""

# ruff: noqa: T201

from __future__ import annotations

import os
import sqlite3
from pathlib import Path

from lemmatizer_be._utils import _fetch_unzip, dir_empty

DATA_DIR = Path(
    os.environ.get(
        "LEMMATIZER_BE_DATA_DIR",
        Path(
            "~",
            ".alerus",
            "shared",
        ),
    ),
    "lemma_data",
).expanduser()
DATA_DIR.mkdir(parents=True, exist_ok=True)

LEMMA_DATA_URL = "https://github.com/alex-rusakevich/lemmatizer-be/releases/latest/download/lemma_data.zip"


class BnkorpusLemmatizer:
    """Belarusian language lemmatizer based on bnkorpus."""

    def __init__(self):
        """Connect to the lemma data sqlite3 db."""
        if dir_empty(DATA_DIR):
            print("The lemmatizer's data is missing, downloading...")
            _fetch_unzip(LEMMA_DATA_URL, DATA_DIR)
            print("The lemmatizer's data has been downloaded successfully.")

        self._conn = sqlite3.connect(str(DATA_DIR / "lemma_data.sqlite3"), check_same_thread=False)
        # check_same_thread=False because we only read, but not change anything
        self._cursor = self._conn.cursor()

    def lemmas(self, word: str, pos: str | None = None) -> list[str]:
        """Return list of all the lemmas for the word.

        Parameters
        ----------
        word : str
            the word lemmatizer finds lemmas for

        pos : Optional[str]
            part of speech letter, see https://bnkorpus.info/grammar.be.html.
            ``None`` means "select all".

        Returns
        -------
        list[str]
            list of lemmas if any

        """
        self._cursor.execute("SELECT * FROM lemma_data WHERE form = ?", (word,))
        result = self._cursor.fetchone()
        lemmas = []

        if result:
            lemmas = result[1].split(";")

        searched_pos = pos

        if isinstance(pos, str):
            searched_pos = pos.upper()

        filtered_lemmas = []

        for lemma in lemmas:
            lemma_text, lemma_pos = lemma.split("|")

            if searched_pos is not None and lemma_pos != searched_pos:
                continue

            if not lemma_text:
                lemma_text = word

            filtered_lemmas.append(lemma_text)

        return list(set(filtered_lemmas))

    def lemmatize(self, word: str, pos: str | None = None) -> str:
        """Lemmatize ``word`` by picking the shortest of the possible lemmas.

        Uses ``self.lemmas()`` internally.
        Returns the input word unchanged if it cannot be found in WordNet.

        Parameters
        ----------
        word : str
            the word lemmatizer finds lemma for

        pos : Optional[str]
            part of speech letter, see https://bnkorpus.info/grammar.be.html.
            ``None`` means "select all".

        Returns
        -------
        str
            the lemma found by lemmatizer

        """
        lemmas = self.lemmas(word, pos=pos)
        return min(lemmas, key=len) if lemmas else word



# ruff: noqa: T201

import timeit

from lemmatizer_be import BnkorpusLemmatizer

if __name__ == "__main__":
    lm = BnkorpusLemmatizer()
    execution_time = timeit.timeit("lm.lemmas('перавырашаць')", globals=globals(), number=100_000)
    print(f"Execution time: {execution_time:.9f} seconds")
